/* Copyright (c) 2014, Giovanni Spina <giovanni@codeforfun.it>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the uploader nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL GIOVANNI SPINA BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/

#include <LCD_Driver.h>
#include <Fonts.h>

#define swap(a, b) { uint8 x = a; a = b; b = x; }

LCDInfo _lcd;

const uint8* _colorMap;

const uint8 RGB12ColorMap[] = 
{
	// number of bytes
	48,

	// red map
	0x00,0x02,0x04,0x06,0x08,0x0A,0x0C,0x0E,0x10,0x12,0x14,0x16,0x18,0x1A,0x1C,0x1F,

	// green map
	0x00,0x07,0x0B,0x0F,0x13,0x17,0x1B,0x1F,0x23,0x27,0x2B,0x2F,0x33,0x37,0x3B,0x3F,

	// blue map
	0x00,0x02,0x04,0x06,0x08,0x0A,0x0C,0x0E,0x10,0x12,0x14,0x16,0x18,0x1A,0x1C,0x1F
};

const uint8 RGB8ColorMap[] = 
{
	// number of bytes
	48,

	// red map
	0, 4, 9, 13, 18, 22, 27, 31, 0, 0, 0, 0, 0, 0, 0, 0,

	// green map
	0, 9, 18, 27, 36, 45, 54, 63, 0, 0, 0, 0, 0, 0, 0, 0,

	// blue map
	0, 10, 21, 31, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

const uint8 RGB8ColorMapEpson[] =
{
	// number of bytes
	20,

	// red map
	0, 2, 4, 6, 9, 11, 13, 15,

	// green map
	0, 2, 4, 6, 9, 11, 13, 15,

	// blue map
	0, 5, 10, 15
};

// Commands
uint8 PASETCMD[] =
{
	PASET,
	PASETP
};

uint8 CASETCMD[] =
{
	CASET,
	CASETP
};

uint8 RAMWRCMD[] =
{
	RAMWR,
	RAMWRP
};

uint8 NOPCMD[] =
{
	NOP,
	NOPP
};

uint8 RGBSETCMD[] =
{
	RGBSET8,
	RGBSET
};

void LCD_Cmd(uint8 cmd)
{
	uint8 i;

	DATA_PIN_DIRECTION = 0;

	CS_PIN = 0;
	DATA_PIN = 0;
	SCK_PIN = 0;

	Delay_us(1);
        
	SCK_PIN = 1;

	for (i = 0; i < 8; ++i)
	{
		if ((cmd & 0x80) == 0x80)
			DATA_PIN = 1;
		else
			DATA_PIN = 0;

		SCK_PIN = 0;
		Delay_us(1);
		SCK_PIN = 1;
		cmd <<= 1;
	}

	CS_PIN = 1;
}

uint8 LCD_Read()
{
	uint8 SPIData, i;
	DATA_PIN = 0;
	DATA_PIN_DIRECTION = 1;

	SPIData = 0;

	for (i = 0; i < 8; i++)
	{
		SPIData <<= 1;
		SCK_PIN = 1;
		SPIData += DATA_PIN;
		SCK_PIN = 0;
	}

	CS_PIN = 1;
	DATA_PIN_DIRECTION = 0;

	return SPIData;
}

uint8 LCD_ReadReg(uint16 reg)
{
	LCD_Cmd(reg);
	return LCD_Read();
}

uint32 LCD_ReadID()
{
	uint8 a, b, c;

	a = LCD_ReadReg(RDID1);
	b = LCD_ReadReg(RDID2);
	c = LCD_ReadReg(RDID3);

	return (((uint32)a) << 16) | (((uint16)b) << 8) | c;
}

void LCD_Data(uint8 Data)
{
	uint8 i;

	DATA_PIN_DIRECTION = 0;

	CS_PIN = 0;
	DATA_PIN = 1;
	SCK_PIN = 0;
        
	Delay_us(1);
        
	SCK_PIN = 1;

	for (i = 0; i < 8; ++i)
	{
		if ((Data & 0x80) == 0x80)
			DATA_PIN = 1;
		else
			DATA_PIN = 0;

		SCK_PIN = 0;
		Delay_us(1);
		SCK_PIN = 1;
		Data <<= 1;
	}

	CS_PIN = 1;
}

void LCD_locateArea(uint8 x, uint8 y, uint8 nRows, uint8 nCols)
{
	if (_lcd.type >= UNKNOWN_CONTROLLER)
		return;

	LCD_Cmd(PASETCMD[_lcd.type]);
	LCD_Data(x);
	LCD_Data(x + nRows);
	LCD_Cmd(ENDPAGE);

	LCD_Cmd(CASETCMD[_lcd.type]);
	LCD_Data(y);
	LCD_Data(y + nCols);
	LCD_Data(ENDCOL);
}

void LCD_locate(uint8 x, uint8 y)
{
	if (_lcd.type >= UNKNOWN_CONTROLLER)
		return;

	LCD_Cmd(PASETCMD[_lcd.type]);
	LCD_Data(x);
	LCD_Cmd(ENDPAGE);

	LCD_Cmd(CASETCMD[_lcd.type]);
	LCD_Data(y);
	LCD_Cmd(ENDCOL);
}

void LCD_WriteRomData(const uint8* RomData, uint16 bytes)
{
	uint16 i;

	for (i = 0; i < bytes; i++)
		LCD_Data(RomData[i]);
}

void LCD_setColMode(uint8 mod, const uint8* table)
{
	if (_lcd.type == EPSON_CONTROLLER)
	{
		// not supported from epson
		if (mod == COLOR16)
			return;

		LCD_Cmd(DATCTL);		// data control(EPSON)
		LCD_Data(0x03);			// correct for normal sin7
		LCD_Data(0x00);			// normal RGB arrangement
		LCD_Data(mod-1);		// GrayScale
	} else {
		LCD_Cmd(COLMOD);
		LCD_Data(mod);
	}

	LCD_Cmd(NOPCMD[_lcd.type]);

	if (table)
	{
		// load the color table
		LCD_Cmd(RGBSETCMD[_lcd.type]);
		LCD_WriteRomData(table + 1, table[0]);
		_colorMap = table;
	}

	_lcd.format = mod;
	LCD_Cmd(NOPCMD[_lcd.type]);
}

void LCD_HardwareReset()
{
	Delay_ms(200);

	SCK_PIN = 0;
	DATA_PIN = 0;
	Delay_us(10);

	CS_PIN = 1;
	Delay_us(10);

	RESET_PIN = 0;
	Delay_ms(200);

	RESET_PIN = 1;
	Delay_ms(200);

	SCK_PIN = 1;
	DATA_PIN = 1;
	Delay_us(10);
}

void LCD_InitPhillips(uint8 colmode) 
{
	// Sleep out
	LCD_Cmd(SLEEPOUT);

	LCD_setColMode(colmode, RGB12ColorMap);

	// Memory access controler
	LCD_Cmd(MADCTL); 
	LCD_Data(0x8);

	// Write contrast
	LCD_Cmd(SETCON); 
	LCD_Data(0x30); // contrast 0x30

	_lcd.contrast = 0x30;

	Delay_ms(1); 

	// Display On
	LCD_Cmd(DISPON);
}

void LCD_InitEpson(uint8 colmode)
{
	LCD_Cmd(DISCTL);			// display control(EPSON)
	LCD_Data(0x0C);				// 12 = 1100 - CL dividing ratio [don't divide] switching period 8H (default)
	LCD_Data(0x20);  
	LCD_Data(0x00);
	LCD_Data(0x01);
        
	LCD_Cmd(COMSCN);			// common scanning direction(EPSON)
	LCD_Data(0x01);
        
	LCD_Cmd(OSCON);				// internal oscialltor ON(EPSON)
        
	LCD_Cmd(SLPOUT);			// sleep out(EPSON)
        
	LCD_Cmd(PWRCTR);			// power ctrl(EPSON)
	LCD_Data(0x0F);				// everything on, no external reference resistors

	LCD_setColMode(colmode, RGB8ColorMapEpson);

	LCD_Cmd(VOLCTR);			// electronic volume, this is the contrast/brightness(EPSON)
	LCD_Data(0x24);				// volume (contrast) setting - fine tuning, original
	LCD_Data(0x03);				// internal resistor ratio - coarse adjustment

	_lcd.contrast = 0x24;

	LCD_Cmd(NOP);				// nop(EPSON)
	Delay_ms(200);
	LCD_Cmd(DISON);				// display on(EPSON)
}

void LCD_Init(uint8 type, uint8 colmode)
{
	uint32 id;

	SCK_PIN = 0;
	DATA_PIN = 0;
	CS_PIN = 1;
	RESET_PIN = 1;
	SCK_PIN_DIRECTION = 0;
	CS_PIN_DIRECTION = 0;
	DATA_PIN_DIRECTION = 0;
	RESET_PIN_DIRECTION = 0;

	LCD_HardwareReset();

	if (type != UNKNOWN_CONTROLLER)
	{
		_lcd.width = 132;
		_lcd.height = 132;
		_lcd.type = type;

		if (_lcd.type == EPSON_CONTROLLER)
			LCD_InitEpson(colmode);
		else LCD_InitPhillips(colmode);

		return;
	}

	id = LCD_ReadID();

	id &= 0xFF;

	_lcd.width = 132;
	_lcd.height = 132;

	switch (id)
	{
	case 0x0E:
	case 0x24:
		if (id == 0x0E)
		{
			_lcd.width = 98;
			_lcd.height = 70;
		}

		_lcd.type = PHILLIPS_CONTROLLER;
		LCD_InitPhillips(colmode);
		break;
	default:
		_lcd.format = COLOR12;
		_lcd.type = EPSON_CONTROLLER;
		LCD_InitEpson(colmode);
		break;
	}
}

void LCD_contrast(int8 contrast)
{
	if (contrast < -64 || contrast > 63) return;

	_lcd.contrast = contrast;

	if (_lcd.type == EPSON_CONTROLLER)
	{
		LCD_Cmd(VOLCTR);
		LCD_Data(contrast);
		LCD_Data(3);
		LCD_Cmd(NOP);
	} else
	{
		LCD_Cmd(SETCON);
		LCD_Data(contrast);
		LCD_Cmd(NOPP);
	}
}

int8 LCD_getContrast()
{
	return _lcd.contrast;
}

void LCD_enable()
{
	if (_lcd.type == EPSON_CONTROLLER)
		LCD_Cmd(DISOFF);
	else
		LCD_Cmd(DISPOFF);
}

void LCD_disable()
{
	if (_lcd.type == EPSON_CONTROLLER)
		LCD_Cmd(DISON);
	else
		LCD_Cmd(DISPON);
}

void LCD_drawPixel(uint8 x, uint8 y, uint16 color)
{
	if (_lcd.type == EPSON_CONTROLLER && _lcd.format == COLOR16)
		return;

	LCD_locate(x, y);
	LCD_Cmd(RAMWRCMD[_lcd.type]);

	if (_lcd.format == COLOR8) {
		LCD_Data((uint8)color);
	} else if (_lcd.format == COLOR12) {
		LCD_Data((color >> 4) & 0xFF);
		LCD_Data(((color & 0xF) << 4) | ((color >> 8) & 0xF));
		LCD_Data(color & 0xFF);
		LCD_Cmd(NOPCMD[_lcd.type]);
	} else {
		LCD_Data((uint8)color>>8);
		LCD_Data(color&0xFF);
	}
}

uint16 LCD_getPixel(uint8 x, uint8 y) 
{
	// Phillips hasn't got a read cmd and not sure if Epson has MISO
	if (_lcd.type == PHILLIPS_CONTROLLER) return 0;

	LCD_locate(x, y);
	LCD_Cmd(RAMRD);

	return (uint16)(LCD_Read() << 8)|LCD_Read();
}

void LCD_drawHorizontalLine(uint8 x, uint8 y, uint8 length, uint16 color)
{
	uint8 i;
	
	if (y >= _lcd.height) return;

	if (x + length > _lcd.width)
		length = _lcd.width - x;

	for (i = 0; i < length; i++)
		LCD_drawPixel(x+i, y, color);
}

void LCD_drawVerticalLine(uint8 x, uint8 y, uint8 length, uint16 color)
{
	uint8 i;
	
	if (x >= _lcd.width) return;

	if (y + length > _lcd.height)
		length = _lcd.height - y;

	for (i = 0; i < length; i++)
		LCD_drawPixel(x, y+i, color);
}

void LCD_drawLine(uint8 p1x, uint8 p1y, uint8 dirx, uint8 diry, uint16 color)
{
	int dx;
	int dy;
	int changeX;
	int changeY;
	int x, y, p;

	if (p1x >= _lcd.width || dirx > _lcd.width)
		return;

	if (p1y > _lcd.height || diry > _lcd.height)
		return;

	dx = dirx - p1x;
	dy = diry - p1y;
	dx = (dx < 0) ? -dx : dx;
	dy = (dy < 0) ? -dy : dy;

	changeX = (dirx < p1x) ? -1 : 1;
	changeY = (diry < p1y) ? -1 : 1;

	x = p1x;
	y = p1y;
	p = 2 * dy - dx;

	LCD_drawPixel(x, y, color);

	if (dy < dx)
	{
		while (x != dirx)
		{
			x = x + changeX;
			if (p < 0)
				p = p + 2 * dy;
			else
			{
				y = y + changeY;
				p = p + 2 * dy - 2 * dx;
			}

			LCD_drawPixel(x, y, color);
		}
	}
	else
	{
		while (y != diry)
		{
			y = y + changeY;
			if (p < 0)
				p = p + 2 * dx;
			else
			{
				x = x + changeX;
				p = p + 2 * dx - 2 * dy;
			}

			LCD_drawPixel(x, y, color);
		}
	}
}

void LCD_drawCircle(uint8 x0, uint8 y0, uint8 radius, uint16 color)
{
	int8 x = radius, y = 0;
	int8 radiusError = 1-x;
 
	while (x >= y)
	{
		LCD_drawPixel(x + x0, y + y0, color);
		LCD_drawPixel(y + x0, x + y0, color);
		LCD_drawPixel(-x + x0, y + y0, color);
		LCD_drawPixel(-y + x0, x + y0, color);
		LCD_drawPixel(-x + x0, -y + y0, color);
		LCD_drawPixel(-y + x0, -x + y0, color);
		LCD_drawPixel(x + x0, -y + y0, color);
		LCD_drawPixel(y + x0, -x + y0, color);
		y++;
		
		if (radiusError<0)
			radiusError += 2 * y + 1;
		else {
			x--;
			radiusError+= 2 * (y - x + 1);
		}
	}
}

void LCD_drawFillCircle(uint8 x0, uint8 y0, uint8 radius, uint16 color)
{
	int8 x = radius, y = 0;
	int8 radiusError = 1-x;
 
	while (x >= y)
	{
		LCD_drawHorizontalLine(-x + x0, y + y0, 2*x, color);
		LCD_drawHorizontalLine(-y + x0, x + y0, 2*y, color);
		LCD_drawHorizontalLine(-x + x0, -y + y0, 2*x, color);
		LCD_drawHorizontalLine(-y + x0, -x + y0, 2*y, color);
		y++;
		
		if (radiusError<0)
			radiusError += 2 * y + 1;
		else {
			x--;
			radiusError+= 2 * (y - x + 1);
		}
	}
}

void LCD_drawRect(uint8 x, uint8 y, uint8 width, uint8 height, uint16 color)
{
	LCD_drawHorizontalLine(x, y, width, color);
	LCD_drawHorizontalLine(x, y + height - 1, width, color);

	LCD_drawVerticalLine(x, y, height, color);
	LCD_drawVerticalLine(x + width - 1, y, height, color);
}

uint16 LCD_NumOfBytes(uint8 w, uint8 h, bool flag)
{
	uint16 bytes;

	switch (_lcd.format)
	{
	case COLOR8:
		bytes = w*h;
		break;
	case COLOR12:
		if (!flag)
			bytes = (uint16)(w*h/2.0)+1;
		else
			bytes = (uint16)(w*h*3.0/2.0)+1;
		break;
	case COLOR16:
		bytes = w*h*2;
		break;
	}

	return bytes;
}

void LCD_drawFillRect(uint8 x, uint8 y, uint8 width, uint8 height, uint16 color)
{
	uint16 i, bytes;

	LCD_locateArea(x, y, width-1, height-1);

	LCD_Cmd(RAMWRCMD[_lcd.type]); 

	bytes = LCD_NumOfBytes(width, height, false);

	for (i = 0; i <= bytes; i++) 
	{ 
		if (_lcd.format == COLOR8)
			LCD_Data((uint8)color);
		else if (_lcd.format == COLOR12) {
			LCD_Data((color>>4)&0x00FF);
			LCD_Data(((color&0x0F)<<4)|(color>>8)&0x0F);
			LCD_Data(color&0x0FF);
		} else {
			LCD_Data((uint8)color>>8);
			LCD_Data((uint8)color&0xFF);
		}
	}

	LCD_Cmd(NOPCMD[_lcd.type]);
}

void LCD_drawTriangle(uint8 p0x, uint8 p0y, uint8 p1x, uint8 p1y, uint8 p2x, uint8 p2y, uint16 color)
{
	LCD_drawLine(p0x, p0y, p1x, p1y, color);
	LCD_drawLine(p1x, p1y, p2x, p2y, color);
	LCD_drawLine(p2x, p2y, p0x, p0y, color);
}

void LCD_drawFilledTriangle(uint8 x0, uint8 y0, uint8 x1, uint8 y1, uint8 x2, uint8 y2, uint16 color)
{
	int8 a, b, y, last;
	int16
		dx01 = x1 - x0,
		dy01 = y1 - y0,
		dx02 = x2 - x0,
		dy02 = y2 - y0,
		dx12 = x2 - x1,
		dy12 = y2 - y1,
		sa = 0,
		sb = 0;

	if (y0 > y1) {
		swap(y0, y1);
		swap(x0, x1);
	}

	if (y1 > y2) {
		swap(y2, y1);
		swap(x2, x1);
	}
	
	if (y0 > y1) {
		swap(y0, y1);
		swap(x0, x1);
	}

	if (y0 == y2) {
		a = b = x0;
		if(x1 < a)
			a = x1;
		else if(x1 > b)
			b = x1;
		if(x2 < a)
			a = x2;
		else if(x2 > b)
			b = x2;
		LCD_drawHorizontalLine(a, y0, b-a+1, color);
		return;
	}

	if (y1 == y2) 
		last = y1;
	else
		last = y1-1;

	for (y = y0; y <= last; y++) {
		a = x0 + sa / dy01;
		b = x0 + sb / dy02;
		sa += dx01;
		sb += dx02;
		if (a > b) swap(a, b);
		LCD_drawHorizontalLine(a, y, b-a+1, color);
	}
	sa = dx12 * (y - y1);
	sb = dx02 * (y - y0);

	for (; y <= y2; y++) {
		a = x1 + sa / dy12;
		b = x0 + sb / dy02;
		sa += dx12;
		sb += dx02;
		if(a > b) swap(a,b);
		LCD_drawHorizontalLine(a, y, b-a+1, color);
	}
}

void LCD_drawImage(const uint8* im, uint8 x, uint8 y)
{
	uint16 bytes;
	uint8 format = _lcd.format;

	LCD_setColMode(im[0], NULL);
	bytes = LCD_NumOfBytes(im[1], im[2], true);

	LCD_locateArea(x, y, im[1]-1, im[2]-1);	
	LCD_Cmd(RAMWRCMD[_lcd.type]);
	LCD_WriteRomData(im+3, bytes);
	LCD_Cmd(NOPCMD[_lcd.type]);

	LCD_setColMode(format, NULL);
}

uint8 LCD_getImageWidth(const uint8* im)
{
	return im[1];
}

uint8 LCD_getImageHeight(const uint8* im)
{
	return im[2];
}

void LCD_clear(uint16 color)
{
	LCD_drawFillRect(0, 0, _lcd.width, _lcd.height, color);
}

text_pos LCD_drawChar(uint8 x, uint8 y, char c, uint16 color, uint16 bcolor, uint8 size)
{
	text_pos pos;

	uint16 i, j, nCols, nRows;
	uint16 Word0, Word1;
	uint8 PixelCol, Mask;
	
	const uint8* pFont;
	const uint8* pChar;
	const uint8* FontTable[] = 
	{
		FONT6x8,
		FONT8x8,
		FONT8x16
	};
  
	pFont = FontTable[size];
	nRows = pFont[0];
	nCols = pFont[1];

	if (c == '\n')
	{
		pos.x = 0;
		pos.y = y + nCols;
		return pos;
	}

	pChar = pFont + 2 + (nCols * (c - 32)); 

	LCD_locateArea(x, y, nRows-1, nCols-1);
	LCD_Cmd(RAMWRCMD[_lcd.type]);

	for (i = 0; i < nCols; i++) 
	{
		PixelCol = pChar[i];	
		Mask = 0x80;

		for (j = 0; j < nRows; j+=2)
		{
			if ((PixelCol & Mask))
				Word0 = color;
			else
				Word0 = bcolor;
			
			Mask = Mask >> 1;
			
			if ((PixelCol & Mask))
				Word1 = color;
			else
				Word1 = bcolor;
			
			Mask = Mask >> 1;

			switch (_lcd.format)
			{
			case COLOR8:
				LCD_Data(Word0);
				LCD_Data(Word1);
				break;
			case COLOR12:
				LCD_Data((Word0 >> 4) & 0xFF); 
				LCD_Data(((Word0 & 0xF) << 4) | ((Word1 >> 8) & 0xF)); 
				LCD_Data(Word1 & 0xFF);
				break;
			default:
				LCD_Data(Word0>>8);
				LCD_Data(Word0&0xFF);
				LCD_Data(Word1>>8);
				LCD_Data(Word1&0xFF);
				break;
			}
		}
	} 

	LCD_Cmd(NOPCMD[_lcd.type]);

	pos.x =  x + nRows;
	pos.y = y;

	return pos;
} 

#ifndef HAS_STRROM
int16 strlen_rom(const char* text)
{
	int16 i;
	for (i = 0; text[i] != '\0'; i++);

	return i;
}
#else
extern int16 strlen_rom(const char* text);
#endif

uint8 LCD_getRomStringWidth(const char* text, uint8 size)
{
	const uint8* pFont;
	const uint8* FontTable[] = 
	{
		FONT6x8,
		FONT8x8,
		FONT8x16
	};

	pFont = FontTable[size];
	
	return pFont[0] * strlen_rom(text);
}

uint8 LCD_getStringWidth(char* text, uint8 size)
{
	const uint8* pFont;
	const uint8* FontTable[] = 
	{
		FONT6x8,
		FONT8x8,
		FONT8x16
	};

	pFont = FontTable[size];
	
	return pFont[0] * strlen(text);
}

void LCD_drawString(uint8 x, uint8 y, char* text, uint16 color, uint16 bcolor, uint8 Size, uint8 justify)
{
	text_pos pos;
	uint8 i;
	uint16 len;

	switch (justify)
	{
	case JUSTIFY_LEFT:
		pos.x = x;
		break;
	case JUSTIFY_CENTER:
		pos.x = _lcd.width / 2.f - LCD_getStringWidth(text, Size) / 2.f;
		break;
	default:
		pos.x = _lcd.width - LCD_getStringWidth(text, Size);
		break;
	}
	pos.y = y;

	len = strlen(text);

	for (i = 0;i <= len; i++) 
	{ 
		pos = LCD_drawChar(pos.x, pos.y, *text++, color, bcolor, Size); 

		if (!pos.x)
			pos.x = x;
		else if (pos.x >= _lcd.width)
			break;
		
		if (pos.y >= _lcd.height)
			break;
	} 
}

void LCD_drawRomString(uint8 x, uint8 y, const char* text, uint16 color, uint16 bcolor, uint8 Size, uint8 justify)
{
	text_pos pos;
	uint8 i;
	uint16 len;

	switch (justify)
	{
	case JUSTIFY_LEFT:
		pos.x = x;
		break;
	case JUSTIFY_CENTER:
		pos.x = _lcd.width / 2.f - LCD_getRomStringWidth(text, Size) / 2.f;
		break;
	default:
		pos.x = _lcd.width - LCD_getRomStringWidth(text, Size);
		break;
	}
	pos.y = y;

	len = strlen_rom(text);

	for (i = 0; i < len; i++) 
	{ 
		pos = LCD_drawChar(pos.x, pos.y, *text++, color, bcolor, Size); 

		if (!pos.x)
			pos.x = x;
		else if (pos.x >= _lcd.width)
			break;
		
		if (pos.y >= _lcd.height)
			break;
	}
}

void LCD_getScreenSize(uint8* w, uint8* h)
{
	if (w)
		*w = _lcd.width;
	
	if (h)
		*h = _lcd.height;
}